import logging
from typing import AsyncGenerator, Generator
from uuid import NAMESPACE_OID, uuid5

import parso
import tiktoken

from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.infrastructure.engine import DataPoint
from cognee.shared.CodeGraphEntities import CodeFile, CodePart, SourceCodeChunk

logger = logging.getLogger(__name__)


def _count_tokens(tokenizer: tiktoken.Encoding, source_code: str) -> int:
    return len(tokenizer.encode(source_code))


def _get_naive_subchunk_token_counts(
    tokenizer: tiktoken.Encoding, source_code: str, max_subchunk_tokens: int = 8000
) -> list[tuple[str, int]]:
    """Splits source code into subchunks of up to max_subchunk_tokens and counts tokens."""

    token_ids = tokenizer.encode(source_code)
    subchunk_token_counts = []

    for start_idx in range(0, len(token_ids), max_subchunk_tokens):
        subchunk_token_ids = token_ids[start_idx : start_idx + max_subchunk_tokens]
        token_count = len(subchunk_token_ids)
        subchunk = "".join(
            tokenizer.decode_single_token_bytes(token_id).decode("utf-8", errors="replace")
            for token_id in subchunk_token_ids
        )
        subchunk_token_counts.append((subchunk, token_count))

    return subchunk_token_counts


def _get_subchunk_token_counts(
    tokenizer: tiktoken.Encoding,
    source_code: str,
    max_subchunk_tokens: int = 8000,
    depth: int = 0,
    max_depth: int = 100,
) -> list[tuple[str, int]]:
    """Splits source code into subchunk and counts tokens for each subchunk."""
    if depth > max_depth:
        return _get_naive_subchunk_token_counts(tokenizer, source_code, max_subchunk_tokens)

    try:
        module = parso.parse(source_code)
    except Exception as e:
        logger.error(f"Error parsing source code: {e}")
        return []

    if not module.children:
        logger.warning("Parsed module has no children (empty or invalid source code).")
        return []

    # Handle cases with only one real child and an EndMarker to prevent infinite recursion.
    if len(module.children) <= 2:
        module = module.children[0]

    subchunk_token_counts = []
    for child in module.children:
        subchunk = child.get_code()
        token_count = _count_tokens(tokenizer, subchunk)

        if token_count == 0:
            continue

        if token_count <= max_subchunk_tokens:
            subchunk_token_counts.append((subchunk, token_count))
            continue

        if child.type == "string":
            subchunk_token_counts.extend(
                _get_naive_subchunk_token_counts(tokenizer, subchunk, max_subchunk_tokens)
            )
            continue

        subchunk_token_counts.extend(
            _get_subchunk_token_counts(
                tokenizer, subchunk, max_subchunk_tokens, depth=depth + 1, max_depth=max_depth
            )
        )

    return subchunk_token_counts


def _get_chunk_source_code(
    code_token_counts: list[tuple[str, int]], overlap: float, max_tokens: int
) -> tuple[list[tuple[str, int]], str]:
    """Generates a chunk of source code from tokenized subchunks with overlap handling."""
    current_count = 0
    cumulative_counts = []
    current_source_code = ""

    for i, (child_code, token_count) in enumerate(code_token_counts):
        current_count += token_count
        cumulative_counts.append(current_count)
        if current_count > max_tokens:
            break
        current_source_code += f"\n{child_code}"

    if current_count <= max_tokens:
        return [], current_source_code.strip()

    cutoff = 1
    for i, cum_count in enumerate(cumulative_counts):
        if cum_count > (1 - overlap) * max_tokens:
            break
        cutoff = i

    return code_token_counts[cutoff:], current_source_code.strip()


def get_source_code_chunks_from_code_part(
    code_file_part: CodePart,
    max_tokens: int = 8192,
    overlap: float = 0.25,
    granularity: float = 0.1,
    model_name: str = "text-embedding-3-large",
) -> Generator[SourceCodeChunk, None, None]:
    """Yields source code chunks from a CodePart object, with configurable token limits and overlap."""
    if not code_file_part.source_code:
        logger.error(f"No source code in CodeFile {code_file_part.id}")
        return

    vector_engine = get_vector_engine()
    embedding_model = vector_engine.embedding_engine.model
    model_name = embedding_model.split("/")[-1]
    tokenizer = tiktoken.encoding_for_model(model_name)
    max_subchunk_tokens = max(1, int(granularity * max_tokens))
    subchunk_token_counts = _get_subchunk_token_counts(
        tokenizer, code_file_part.source_code, max_subchunk_tokens
    )

    previous_chunk = None
    while subchunk_token_counts:
        subchunk_token_counts, chunk_source_code = _get_chunk_source_code(
            subchunk_token_counts, overlap, max_tokens
        )
        if not chunk_source_code:
            continue
        current_chunk = SourceCodeChunk(
            id=uuid5(NAMESPACE_OID, chunk_source_code),
            code_chunk_of=code_file_part,
            source_code=chunk_source_code,
            previous_chunk=previous_chunk,
        )
        yield current_chunk
        previous_chunk = current_chunk


async def get_source_code_chunks(
    data_points: list[DataPoint],
) -> AsyncGenerator[list[DataPoint], None]:
    """Processes code graph datapoints, create SourceCodeChink datapoints."""
    # TODO: Add support for other embedding models, with max_token mapping
    for data_point in data_points:
        try:
            yield data_point
            if not isinstance(data_point, CodeFile):
                continue
            if not data_point.contains:
                logger.warning(f"CodeFile {data_point.id} contains no code parts")
                continue
            for code_part in data_point.contains:
                try:
                    yield code_part
                    for source_code_chunk in get_source_code_chunks_from_code_part(code_part):
                        yield source_code_chunk
                except Exception as e:
                    logger.error(f"Error processing code part: {e}")
        except Exception as e:
            logger.error(f"Error processing data point: {e}")



from cognee.exceptions import CogneeApiError
from fastapi import status


class GroupNotFoundError(CogneeApiError):
    """User group not found"""

    def __init__(
        self,
        message: str = "User group not found.",
        name: str = "GroupNotFoundError",
        status_code=status.HTTP_404_NOT_FOUND,
    ):
        super().__init__(message, name, status_code)


class UserNotFoundError(CogneeApiError):
    """User not found"""

    def __init__(
        self,
        message: str = "No user found in the system. Please create a user.",
        name: str = "UserNotFoundError",
        status_code=status.HTTP_404_NOT_FOUND,
    ):
        super().__init__(message, name, status_code)


class PermissionDeniedError(CogneeApiError):
    def __init__(
        self,
        message: str = "User does not have permission on documents.",
        name: str = "PermissionDeniedError",
        status_code=status.HTTP_403_FORBIDDEN,
    ):
        super().__init__(message, name, status_code)



import asyncio
import os
from concurrent.futures import ProcessPoolExecutor
from typing import AsyncGenerator
from uuid import NAMESPACE_OID, uuid5

import aiofiles

from cognee.shared.CodeGraphEntities import CodeFile, Repository
from cognee.tasks.repo_processor.get_local_dependencies import get_local_script_dependencies


async def get_py_path_and_source(file_path):
    try:
        async with aiofiles.open(file_path, "r", encoding="utf-8") as f:
            source_code = await f.read()
        return file_path, source_code
    except Exception as e:
        print(f"Error reading file {file_path}: {e}")
        return file_path, None


async def get_py_files_dict(repo_path):
    """Get .py files and their source code"""
    if not os.path.exists(repo_path):
        return {}

    py_files_paths = (
        os.path.join(root, file)
        for root, _, files in os.walk(repo_path)
        for file in files
        if file.endswith(".py")
    )

    py_files_dict = {}
    for file_path in py_files_paths:
        absolute_path = os.path.abspath(file_path)

        if os.path.getsize(absolute_path) == 0:
            continue

        relative_path, source_code = await get_py_path_and_source(absolute_path)
        py_files_dict[relative_path] = {"source_code": source_code}

    return py_files_dict


def get_edge(
    file_path: str, dependency: str, repo_path: str, relative_paths: bool = False
) -> tuple:
    if relative_paths:
        file_path = os.path.relpath(file_path, repo_path)
        dependency = os.path.relpath(dependency, repo_path)
    return (file_path, dependency, {"relation": "depends_directly_on"})


def run_coroutine(coroutine_func, *args, **kwargs):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    result = loop.run_until_complete(coroutine_func(*args, **kwargs))
    loop.close()
    return result


async def get_repo_file_dependencies(repo_path: str) -> AsyncGenerator[list, None]:
    """Generate a dependency graph for Python files in the given repository path."""

    if not os.path.exists(repo_path):
        raise FileNotFoundError(f"Repository path {repo_path} does not exist.")

    py_files_dict = await get_py_files_dict(repo_path)

    repo = Repository(
        id=uuid5(NAMESPACE_OID, repo_path),
        path=repo_path,
    )

    yield [repo]

    with ProcessPoolExecutor(max_workers=12) as executor:
        loop = asyncio.get_event_loop()

        tasks = [
            loop.run_in_executor(
                executor,
                run_coroutine,
                get_local_script_dependencies,
                os.path.join(repo_path, file_path),
                repo_path,
            )
            for file_path, metadata in py_files_dict.items()
            if metadata.get("source_code") is not None
        ]

        results = await asyncio.gather(*tasks)

        code_files = []
        for (file_path, metadata), dependencies in zip(py_files_dict.items(), results):
            source_code = metadata.get("source_code")

            code_files.append(
                CodeFile(
                    id=uuid5(NAMESPACE_OID, file_path),
                    source_code=source_code,
                    extracted_id=file_path,
                    part_of=repo,
                    depends_on=[
                        CodeFile(
                            id=uuid5(NAMESPACE_OID, dependency),
                            extracted_id=dependency,
                            part_of=repo,
                            source_code=py_files_dict.get(dependency, {}).get("source_code"),
                        )
                        for dependency in dependencies
                    ]
                    if dependencies
                    else None,
                )
            )

        yield code_files



import logging
from typing import Dict, List, Optional
from uuid import UUID

from qdrant_client import AsyncQdrantClient, models

from cognee.exceptions import InvalidValueError
from cognee.infrastructure.databases.vector.models.ScoredResult import ScoredResult
from cognee.infrastructure.engine import DataPoint

from ..embeddings.EmbeddingEngine import EmbeddingEngine
from ..vector_db_interface import VectorDBInterface

logger = logging.getLogger("QDrantAdapter")


class IndexSchema(DataPoint):
    text: str

    _metadata: dict = {"index_fields": ["text"], "type": "IndexSchema"}


# class CollectionConfig(BaseModel, extra = "forbid"):
#     vector_config: Dict[str, models.VectorParams] = Field(..., description="Vectors configuration" )
#     hnsw_config: Optional[models.HnswConfig] = Field(default = None, description="HNSW vector index configuration")
#     optimizers_config: Optional[models.OptimizersConfig] = Field(default = None, description="Optimizers configuration")
#     quantization_config: Optional[models.QuantizationConfig] = Field(default = None, description="Quantization configuration")


def create_hnsw_config(hnsw_config: Dict):
    if hnsw_config is not None:
        return models.HnswConfig()
    return None


def create_optimizers_config(optimizers_config: Dict):
    if optimizers_config is not None:
        return models.OptimizersConfig()
    return None


def create_quantization_config(quantization_config: Dict):
    if quantization_config is not None:
        return models.QuantizationConfig()
    return None


class QDrantAdapter(VectorDBInterface):
    name = "Qdrant"
    url: str = None
    api_key: str = None
    qdrant_path: str = None

    def __init__(self, url, api_key, embedding_engine: EmbeddingEngine, qdrant_path=None):
        self.embedding_engine = embedding_engine

        if qdrant_path is not None:
            self.qdrant_path = qdrant_path
        else:
            self.url = url
            self.api_key = api_key

    def get_qdrant_client(self) -> AsyncQdrantClient:
        if self.qdrant_path is not None:
            return AsyncQdrantClient(path=self.qdrant_path, port=6333)
        elif self.url is not None:
            return AsyncQdrantClient(url=self.url, api_key=self.api_key, port=6333)

        return AsyncQdrantClient(location=":memory:")

    async def embed_data(self, data: List[str]) -> List[float]:
        return await self.embedding_engine.embed_text(data)

    async def has_collection(self, collection_name: str) -> bool:
        client = self.get_qdrant_client()
        result = await client.collection_exists(collection_name)
        await client.close()
        return result

    async def create_collection(
        self,
        collection_name: str,
        payload_schema=None,
    ):
        client = self.get_qdrant_client()

        if not await client.collection_exists(collection_name):
            await client.create_collection(
                collection_name=collection_name,
                vectors_config={
                    "text": models.VectorParams(
                        size=self.embedding_engine.get_vector_size(), distance="Cosine"
                    )
                },
            )

        await client.close()

    async def create_data_points(self, collection_name: str, data_points: List[DataPoint]):
        client = self.get_qdrant_client()

        data_vectors = await self.embed_data(
            [DataPoint.get_embeddable_data(data_point) for data_point in data_points]
        )

        def convert_to_qdrant_point(data_point: DataPoint):
            return models.PointStruct(
                id=str(data_point.id),
                payload=data_point.model_dump(),
                vector={"text": data_vectors[data_points.index(data_point)]},
            )

        points = [convert_to_qdrant_point(point) for point in data_points]

        try:
            client.upload_points(collection_name=collection_name, points=points)
        except Exception as error:
            logger.error("Error uploading data points to Qdrant: %s", str(error))
            raise error
        finally:
            await client.close()

    async def create_vector_index(self, index_name: str, index_property_name: str):
        await self.create_collection(f"{index_name}_{index_property_name}")

    async def index_data_points(
        self, index_name: str, index_property_name: str, data_points: list[DataPoint]
    ):
        await self.create_data_points(
            f"{index_name}_{index_property_name}",
            [
                IndexSchema(
                    id=data_point.id,
                    text=getattr(data_point, data_point._metadata["index_fields"][0]),
                )
                for data_point in data_points
            ],
        )

    async def retrieve(self, collection_name: str, data_point_ids: list[str]):
        client = self.get_qdrant_client()
        results = await client.retrieve(collection_name, data_point_ids, with_payload=True)
        await client.close()
        return results

    async def get_distance_from_collection_elements(
        self,
        collection_name: str,
        query_text: str = None,
        query_vector: List[float] = None,
        with_vector: bool = False,
    ) -> List[ScoredResult]:
        if query_text is None and query_vector is None:
            raise ValueError("One of query_text or query_vector must be provided!")

        client = self.get_qdrant_client()

        results = await client.search(
            collection_name=collection_name,
            query_vector=models.NamedVector(
                name="text",
                vector=query_vector
                if query_vector is not None
                else (await self.embed_data([query_text]))[0],
            ),
            with_vectors=with_vector,
        )

        await client.close()

        return [
            ScoredResult(
                id=UUID(result.id),
                payload={
                    **result.payload,
                    "id": UUID(result.id),
                },
                score=1 - result.score,
            )
            for result in results
        ]

    async def search(
        self,
        collection_name: str,
        query_text: Optional[str] = None,
        query_vector: Optional[List[float]] = None,
        limit: int = 5,
        with_vector: bool = False,
    ):
        if query_text is None and query_vector is None:
            raise InvalidValueError(message="One of query_text or query_vector must be provided!")

        client = self.get_qdrant_client()

        results = await client.search(
            collection_name=collection_name,
            query_vector=models.NamedVector(
                name="text",
                vector=query_vector
                if query_vector is not None
                else (await self.embed_data([query_text]))[0],
            ),
            limit=limit,
            with_vectors=with_vector,
        )

        await client.close()

        return [
            ScoredResult(
                id=UUID(result.id),
                payload={
                    **result.payload,
                    "id": UUID(result.id),
                },
                score=1 - result.score,
            )
            for result in results
        ]

    async def batch_search(
        self,
        collection_name: str,
        query_texts: List[str],
        limit: int = None,
        with_vectors: bool = False,
    ):
        """
        Perform batch search in a Qdrant collection with dynamic search requests.

        Args:
        - collection_name (str): Name of the collection to search in.
        - query_texts (List[str]): List of query texts to search for.
        - limit (int): List of result limits for search requests.
        - with_vectors (bool, optional): Bool indicating whether to return vectors for search requests.

        Returns:
        - results: The search results from Qdrant.
        """

        vectors = await self.embed_data(query_texts)

        # Generate dynamic search requests based on the provided embeddings
        requests = [
            models.SearchRequest(
                vector=models.NamedVector(name="text", vector=vector),
                limit=limit,
                with_vector=with_vectors,
            )
            for vector in vectors
        ]

        client = self.get_qdrant_client()

        # Perform batch search with the dynamically generated requests
        results = await client.search_batch(collection_name=collection_name, requests=requests)

        await client.close()

        return [filter(lambda result: result.score > 0.9, result_group) for result_group in results]

    async def delete_data_points(self, collection_name: str, data_point_ids: list[str]):
        client = self.get_qdrant_client()
        results = await client.delete(collection_name, data_point_ids)
        return results

    async def prune(self):
        client = self.get_qdrant_client()

        response = await client.get_collections()

        for collection in response.collections:
            await client.delete_collection(collection.name)

        await client.close()



from uuid import UUID
from typing import Optional
from cognee.infrastructure.databases.relational import get_relational_engine
from ..exceptions import UnauthorizedDataAccessError
from ..models import Data


async def get_data(user_id: UUID, data_id: UUID) -> Optional[Data]:
    """Retrieve data by ID.

    Args:
        user_id (UUID): user ID
        data_id (UUID): ID of the data to retrieve

    Returns:
        Optional[Data]: The requested data object if found, None otherwise
    """
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        data = await session.get(Data, data_id)

        if data and data.owner_id != user_id:
            raise UnauthorizedDataAccessError(
                message=f"User {user_id} is not authorized to access data {data_id}"
            )

        return data
